IdeSourceMap {
  padding-left: 13px;
  padding-right: 13px;
}
IdeSourceMap:dir(ltr) {
  border-left: 1px solid color-mix(in srgb, var(--border-color) 30%, transparent);
}
IdeSourceMap:dir(rtl) {
  border-right: 1px solid color-mix(in srgb, var(--border-color) 30%, transparent);
}
IdeSourceMap textview.GtkSourceMap {
  background: transparent;
  font-family: BuilderBlocks;
  font-size: 1.75pt;
  line-height: 5px;
  margin: 0;
}
IdeSourceMap textview.GtkSourceMap slider {
  padding-left: 13px;
  padding-right: 13px;
  margin-left: -13px;
  margin-right: -13px;
}
IdeSourceMap textview.GtkSourceMap border,
IdeSourceMap textview.GtkSourceMap gutter {
  background: transparent;
}

.hover-display-row:not(:last-child) {
  border-bottom: 1px solid var(--border-color);
}

.hover-display-row {
  padding: 6px;
  padding-left: 12px;
  padding-right: 12px;
}

GtkSourceAssistant.hover-assistant {
  background: var(--popover-bg-color);
  color: var(--popover-fg-color);
  box-shadow: 0 1px 5px 1px rgb(0 0 0 / 9%),
              0 2px 14px 3px rgb(0 0 0 / 5%),
              /* give a bit more pointer space to avoid flickering display */
              0 0 20px 20px transparent;
}
