vte-terminal
{
  padding-top: 7px;
  padding-bottom: 6px;
  padding-left: 6px;
  padding-right: 6px;
}

vte-terminal revealer.size label
{
  background-color: mix(@theme_bg_color, @theme_fg_color, .1);
  padding: 3px 12px;
  border-radius: 8px;
  margin: 4px 4px 0 4px;
}
