.install-progress {
  background-image: linear-gradient(to top,
                                    var(--accent-bg-color) 2px,
                                    rgb(from var(--accent-bg-color) r g b / 0) 2px);
  background-repeat: no-repeat;
  background-position: 0 bottom;
  transition: none;
  background-size: 0;
}
.install-progress:dir(rtl) {
  background-position: 100% bottom;
}
