GbpManualsPathbar {
  padding: 0;
}

GbpManualsPathbar pathbutton button {
  padding: 0px 9px;
  margin: 1px 3px;
}

GbpManualsPathbar pathbutton:not(.leaf) button > * {
  opacity: var(--dim-opacity);
}

GbpManualsPanel tag {
  border-radius: 33px;
  padding: 1px 8px;
  font-size: .8333em;
  font-weight: normal;
}

GbpManualsPanel tag.since {
  background: var(--blue-3);
  color: white;
}

GbpManualsPanel tag.deprecated {
  background: var(--red-2);
  color: white;
}

GbpManualsPanel tag.stability {
  background: var(--yellow-5);
  color: black;
}
