window.workspace headerbar {
  box-shadow: none;
}
window.workspace tabbar {
  background: var(--headerbar-bg-color);
}
window.workspace:backdrop tabbar {
  background: var(--headerbar-backdrop-color);
}
window.workspace:backdrop paneldockchild.center tabbar {
  background: var(--headerbar-bg-color);
}
window.workspace paneldockchild.bottom,
window.workspace paneldockchild.center {
  --headerbar-bg-color: var(--view-bg-color);
  --headerbar-fg-color: var(--view-fg-color);
}
paneldockchild.center tabbar {
  background: var(--view-bg-color);
}
paneldockchild.center headerbar {
  background: var(--view-bg-color);
}
paneldockchild.center windowhandle {
  background: transparent;
}

stylevariantpreview widget.wallpaper {
  border-radius: 6px;
}

.global-search search {
  padding: 12px;
  border-bottom: 1px solid var(--border-color);
}
.global-search search > image {
  -gtk-icon-size: 24px;
}
.global-search search > text {
  font-size: 1.66666em;
  margin-left: 6px;
  margin-right: 6px;
}
.global-search search > text placeholder {
  opacity: 0.55;
}
.global-search search > button.close {
  padding: 2px;
  min-width: 24px;
  min-height: 24px;
  margin-right: 6px;
}
.global-search .preview-bin {
  border-left: 1px solid color-mix(in srgb, var(--border-color) 25%, transparent);
}

notificationstack button {
  margin: 0;
  padding: 0;
  background: none;
}

button.pill.small {
  font-size: .83333em;
  border-radius: 99px;
  margin: 0;
  padding: 1px 12px;
}

/* Styling for search bars */
.searchbar {
  background-color: var(--popover-bg-color);
  background-clip: padding-box;
  color: var(--popover-fg-color);
  padding: 6px;
  border-style: solid;
  border-color: rgb(0 0 0 / 14%);
  border-left-width: 1px;
  border-right-width: 1px;
  border-bottom-width: 1px;
  border-radius: 0 0 12px 12px;
  margin: 0 10px 20px 10px;
  box-shadow: 0 1px 5px 1px rgb(0 0 0 / 9%),
              0 2px 14px 3px rgb(0 0 0 / 5%);
}
.searchbar button.flat.circular {
  min-width: 24px;
  min-height: 24px;
  margin: 0px;
  padding: 3px;
}

/* Images for checkbuttons in various UI */
.checkimage radio,
.checkimage check {
  background: none;
  outline: none;
  box-shadow: none;
  border: none;
  color: var(--window-fg-color);
  -gtk-icon-source: none;
  -gtk-icon-size: 18px;
}
.checkimage radio:checked,
.checkimage check:checked {
  -gtk-icon-source: -gtk-icontheme('object-select-symbolic');
}

/* Statusbar style to match sidebar entries */
panelstatusbar {
  min-height: 28px;
}

/* Sidebar style */
entry.statusbar {
  min-height: 29px;
  font-size: 0.9em;
  box-shadow: inset 0 1px var(--sidebar-border-color);
  border-radius: 0;
}
paneldockchild.start panelframe:last-child entry.statusbar:dir(ltr),
paneldockchild.end panelframe:last-child entry.statusbar:dir(rtl) {
  border-bottom-left-radius: var(--window-radius);
}
paneldockchild.start panelframe:last-child entry.statusbar:dir(rtl),
paneldockchild.end panelframe:last-child entry.statusbar:dir(ltr) {
  border-bottom-right-radius: var(--window-radius);
}
paneldockchild.start,
paneldockchild.end {
  background: var(--sidebar-bg-color);
  color: var(--sidebar-fg-color);
}
paneldockchild.start:backdrop,
paneldockchild.end:backdrop {
  background: var(--sidebar-backdrop-color);
  transition: 200ms ease-in-out;
}
paneldockchild.start .frameheader,
paneldockchild.end .frameheader {
  border-top: 1px solid transparent;
}
paneldockchild.start panelresizer > widget > separator,
paneldockchild.end panelresizer > widget > separator {
  background-color: var(--sidebar-shade-color);
}

/* Make statuspage match the "view" sytle of the center column */
window.workspace paneldockchild.center statuspage {
 background: var(--headerbar-bg-color);
}
window.workspace paneldockchild.bottom {
  background: var(--view-bg-color);
}
