gbpgreppanel {
  background-color: var(--window-bg-color);
}
gbpgreppanel button {
  min-height: 0px;
}
gbpgreppanel {
  border-left: 1px solid color-mix(in srgb, var(--border-color) 10%, transparent);
}
